/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.utils;

import static org.eclipse.emf.ecore.util.EcoreUtil.generateUUID;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtils.createReuseHashAsString;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtils.getReuseElementName;

import java.util.Date;
import java.util.Random;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.ext.reuse.model.AF3ReuseFactory;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;

/**
 * This class provides the creation methods for {@link ReuseLibrary} and related
 * elements.
 *
 * @author bergemann
 */
public class ReuseLibraryModelElementFactory {

	/** Generates a UUID that is usable for {@link ReuseLibrary}. */
	private static String generateReuseUUID() {
		String uuid = generateUUID();
		// "-" might be used later as separator --> remove it from ID
		Random randomStream = new Random();
		char replacementCharacter = (char) (randomStream.nextInt(26) + 'a');
		return uuid.replace('-', replacementCharacter);
	}

	/**
	 * Creates a {@link ReuseLibrary}.
	 *
	 * @return The created reuse library
	 */
	public static ReuseLibrary createReuseLibrary() {
		return createReuseLibrary("Reuse Library");
	}

	/**
	 * Creates a {@link ReuseLibrary} with a given name.
	 *
	 * @param libraryName The name for the new reuse library
	 * @return The created reuse library
	 */
	public static ReuseLibrary createReuseLibrary(String libraryName) {
		ReuseLibrary lib = AF3ReuseFactory.eINSTANCE.createReuseLibrary();
		lib.setName(libraryName);
		lib.setLibraryUUID(generateReuseUUID());
		return lib;
	}

	/**
	 * Creates a {@link ReuseElementSpec} for the given element based on the given
	 * {@link ReuseLibrary} as source.
	 *
	 * @param element       The connected reuse element
	 * @param sourceLibrary The connected reuse library (source of the reuse
	 *                      element)
	 * @return The created reuse specification
	 */
	public static ReuseElementSpec createReuseElementSpec(EObject element, ReuseLibrary sourceLibrary) {
		Date now = new Date();
		ReuseElementSpec spec = AF3ReuseFactory.eINSTANCE.createReuseElementSpec();
		spec.setElementUUID(generateReuseUUID());
		spec.setElementName(getReuseElementName(element));
		spec.setElementHash(createReuseHashAsString(element, now));
		spec.setSourceLibUUID(sourceLibrary.getLibraryUUID());
		spec.setSourceLibName(sourceLibrary.getName());
		spec.setLastUpdate(now);
		return spec;
	}

}
