/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.utils;

import static org.eclipse.emf.ecore.util.EcoreUtil.replace;
import static org.fortiss.tooling.ext.reuse.ToolingReuseActivator.getDefault;
import static org.fortiss.tooling.ext.reuse.service.ReuseProviderService.getPossibleExternalReferenceClasses;
import static org.fortiss.tooling.ext.reuse.service.ReuseProviderService.getReuseElementClasses;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.deleteReuseLibraryFolder;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.getAllSeparatelyStoredReuseLibraries;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.saveReuseLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryModelElementFactory.createReuseElementSpec;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryModelElementFactory.createReuseLibrary;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.showError;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.generateMissingIDs;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.getLargestID;
import static org.fortiss.tooling.kernel.utils.UniqueIDUtils.removeDuplicateIds;

import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.service.IReuseProviderService;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Utility methods for {@link ReuseLibrary}s and related reuse elements and
 * actions.
 *
 * @author bergemann
 */
public class ReuseLibraryUtils {

	/**
	 * Fixes the IDs of the model containing the given {@link EObject}, since
	 * updating reuse elements can produce duplicated IDs.
	 *
	 * @param element A reuse element of the model
	 */
	public static void fixIds(EObject element) {
		EObject root = element;
		while (root.eContainer() != null) {
			root = root.eContainer();
		}
		removeDuplicateIds(root);
		generateMissingIDs(root, getLargestID(root));
	}

	/**
	 * Returns whether the given element can be used as reuse element.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element can be used as reuse element otherwise
	 *         false
	 */
	public static boolean hasReusePossibility(EObject element) {
		for (Class<? extends EObject> cls : getReuseElementClasses()) {
			if (cls.isInstance(element)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns whether the given element is a reuse element and has at least one
	 * valid link to a {@link ReuseLibrary}.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element is a reuse element, otherwise false
	 */
	public static boolean isCorrectReuseElement(EObject element) {
		// first, check if it is a (possible) reuse element in general
		if (hasReusePossibility(element)) {
			// second, check if it has a library reference
			// (only then it is an actual (active) reuse element)
			if (element instanceof IModelElement) {
				for (IModelElementSpecification spec : ((IModelElement) element).getSpecifications()) {
					if (spec instanceof ReuseElementSpec) {
						String libID = ((ReuseElementSpec) spec).getSourceLibUUID();
						if (libID != null && !libID.isEmpty()) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * Returns whether the given element is a reuse element and has at least one
	 * valid link to a {@link ReuseLibrary} OR is a {@link ReuseLibrary} itself.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element is a reuse element OR a reuse library,
	 *         otherwise false
	 */
	public static boolean isCorrectReuseElementOrLibrary(EObject element) {
		return (element instanceof ReuseLibrary) || isCorrectReuseElement(element);
	}

	/**
	 * Returns display name of given {@link ReuseLibrary} (name and UUID).
	 *
	 * @param library The target reuse library
	 * @return Display name as string
	 */
	public static String getLibraryIDName(ReuseLibrary library) {
		return getLibraryIDNameFormat(library.getLibraryUUID(), library.getName());
	}

	/**
	 * Returns the current format for displaying name and UUID of
	 * {@link ReuseLibrary}s.
	 *
	 * @param libraryUUID The UUID of the target library
	 * @param libraryName The name of the target library
	 * @return Formatted string
	 */
	public static String getLibraryIDNameFormat(String libraryUUID, String libraryName) {
		return libraryName + " (" + libraryUUID + ")";
	}

	/**
	 * Returns the (converted) name of the given reuse element (or
	 * {@link ReuseLibrary}), independent of the type as long as its class was
	 * registered for reuse.
	 *
	 * @param element The target reuse element
	 * @return Name as string
	 */
	public static String getReuseElementName(EObject element) {
		if (element instanceof ReuseLibrary) {
			return ((ReuseLibrary) element).getName();
		}
		String name = IReuseProviderService.getInstance().getReuseElementName(element);
		if (name == null) {
			if (element instanceof INamedElement) {
				name = ((INamedElement) element).getName();
			} else {
				name = element.toString();
			}
		}
		return name;
	}

	/**
	 * Sets the name of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @param name    The new name
	 * @return True if action was successful, otherwise false
	 */
	public static boolean setReuseElementName(EObject element, String name) {
		if (element instanceof ReuseLibrary) {
			((ReuseLibrary) element).setName(name);
			return true;
		}
		return IReuseProviderService.getInstance().setReuseElementName(element, name);
	}

	/**
	 * Returns the comment of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @return The requested comment as string
	 */
	public static String getReuseElementComment(EObject element) {
		String comment = null;
		if (element instanceof ReuseLibrary) {
			comment = ((ReuseLibrary) element).getComment();
		} else {
			comment = IReuseProviderService.getInstance().getReuseElementComment(element);
			if (comment == null) {
				if (element instanceof INamedCommentedElement) {
					comment = ((INamedCommentedElement) element).getComment();
				}
			}
		}
		return (comment == null) ? "" : comment;
	}

	/**
	 * Sets the comment of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @param comment The new comment
	 * @return True if action was successful, otherwise false
	 */
	public static boolean setReuseElementComment(EObject element, String comment) {
		if (element instanceof ReuseLibrary) {
			((ReuseLibrary) element).setComment(comment);
			return true;
		}
		return IReuseProviderService.getInstance().setReuseElementComment(element, comment);
	}

	/**
	 * Returns the first UUID of the given reuse element (or {@link ReuseLibrary})
	 * as long as it has (already) a {@link ReuseElementSpec}. If this is not the
	 * case, an empty {@link String} will be returned. This method is perfect for
	 * reuse elements within reuse libraries, because they have definitely only one
	 * {@link ReuseElementSpec} and thus exactly one UUID (= the first one). Reuse
	 * elements that are already placed in projects can have multiple
	 * {@link ReuseElementSpec}s and thus multiple UUIDs, because they could have
	 * been added to several {@link ReuseLibrary}s.
	 *
	 * @param element The target reuse element
	 * @return The requested UUID as string (or empty string)
	 */
	public static String getFirstReuseElementUUID(EObject element) {
		List<String> allUUIDs = getAllReuseElementUUID(element);
		if (allUUIDs != null && !allUUIDs.isEmpty()) {
			// get(0) is safe due to check above
			return allUUIDs.get(0);
		}
		return "";
	}

	/**
	 * Returns all UUIDs of the given reuse element (or {@link ReuseLibrary}) as
	 * long as it has (already) at least one {@link ReuseElementSpec}. If this is
	 * not the case, an empty list will be returned. This method exists because
	 * reuse elements that are already placed in projects can have multiple
	 * {@link ReuseElementSpec}s and thus multiple UUIDs, because they could have
	 * been added to several {@link ReuseLibrary}s.
	 *
	 * @param element The target reuse element
	 * @return The requested UUID as string (or empty string)
	 */
	public static List<String> getAllReuseElementUUID(EObject element) {
		List<String> allUUIDs = new ArrayList<>();
		if (element instanceof ReuseLibrary) {
			allUUIDs.add(((ReuseLibrary) element).getLibraryUUID());
		} else {
			if (element instanceof IModelElement) {
				List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
				for (ReuseElementSpec spec : specs) {
					allUUIDs.add(spec.getElementUUID());
				}
			}
		}
		return allUUIDs;
	}

	/**
	 * Returns the UUID of the given reuse element that is stored in the
	 * {@link ReuseElementSpec} which is connected to the given
	 * {@link ReuseLibrary}. If such a {@link ReuseElementSpec} odes not exist, an
	 * empty {@link String} will be returned.
	 *
	 * @param element The target reuse element
	 * @param library The reuse library to which the reuse UUID of the element
	 *                should belong
	 * @return The requested UUID as string (or empty string)
	 */
	public static String getReuseElementUUIDForLibrary(EObject element, ReuseLibrary library) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec spec : specs) {
				if (spec.getSourceLibUUID().equals(library.getLibraryUUID())) {
					return spec.getElementUUID();
				}
			}
		}
		return "";
	}

	/**
	 * Returns the hash code string of the given reuse element for its origin inside
	 * a specific (given) {@link ReuseLibrary}. If it cannot be found, an empty
	 * {@link String} will be returned.
	 *
	 * @param element       The target reuse element
	 * @param sourceLibrary The library of the original reuse element for the hash
	 * @return The requested hash code as string (or empty string)
	 */
	public static String getReuseElementHashForLibrary(EObject element, ReuseLibrary sourceLibrary) {
		String libraryUuid = sourceLibrary.getLibraryUUID();
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec spec : specs) {
				String referenceLibraryUuid = spec.getSourceLibUUID();
				if (referenceLibraryUuid.equals(libraryUuid)) {
					return spec.getElementHash();
				}
			}
		}
		return "";
	}

	/**
	 * Returns the Last Update Date of the given reuse element as {link Date} as
	 * long as it has (already) a {@link ReuseElementSpec}. If this is not the case,
	 * null will be returned.
	 *
	 * @param element The target reuse element
	 * @return The requested (update) date (or null)
	 */
	public static Date getReuseElementLastUpdate(EObject element) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			if (specs != null && !specs.isEmpty()) {
				// get(0) is safe due to check above
				return specs.get(0).getLastUpdate();
			}
		}
		return null;
	}

	/**
	 * Returns the Last Update Date of the given reuse element as formatted {link
	 * String} as long as it has (already) a {@link ReuseElementSpec}. If this is
	 * not the case, an empty {@link String} will be returned.
	 *
	 * @param element The target reuse element
	 * @return The requested (update) date as string (or empty string)
	 */
	public static String getReuseElementLastUpdateString(EObject element) {
		DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		Date updateDate = getReuseElementLastUpdate(element);
		if (updateDate != null) {
			return dateFormat.format(updateDate);
		}
		return "";
	}

	/**
	 * Sets the Last Update Date of the given reuse element with the given {link
	 * Date} as long as it has (already) a {@link ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 * @param newDate The new (update) date
	 */
	public static void setReuseElementLastUpdate(EObject element, Date newDate) {
		if (element instanceof IModelElement) {
			for (ReuseElementSpec spec : getAllReuseSpecs((IModelElement) element)) {
				spec.setLastUpdate(newDate);
			}
		}
	}

	/**
	 * Sets the Last Update Date of the given reuse element with the current {link
	 * Date} as long as it has (already) a {@link ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 */
	public static void setReuseElementLastUpdateNow(EObject element) {
		setReuseElementLastUpdate(element, new Date());
	}

	/**
	 * Returns all {@link ReuseElementSpec}s of the given element to a
	 * {@link ReuseLibrary}. If none exist, an empty list will be returned.
	 *
	 * @param element The target reuse element
	 * @return A list of all found reuse specifications (or empty list)
	 */
	public static List<ReuseElementSpec> getAllReuseSpecs(IModelElement element) {
		List<ReuseElementSpec> specs = new ArrayList<>();
		for (IModelElementSpecification spec : element.getSpecifications()) {
			if (spec instanceof ReuseElementSpec) {
				// (cast is save due to instance check)
				specs.add((ReuseElementSpec) spec);
			}
		}
		return specs;
	}

	/**
	 * Adds a {@link ReuseElementSpec} to the given element pointing to the given
	 * {@link ReuseLibrary} (as source).
	 *
	 * @param element The target reuse element
	 * @param library The connected reuse library
	 * @return The created reuse specification
	 */
	public static ReuseElementSpec addReuseSpecToElement(IModelElement element, ReuseLibrary library) {
		ReuseElementSpec spec = createReuseElementSpec(element, library);
		element.getSpecifications().add(spec);
		return spec;
	}

	/**
	 * Setups a given reuse element for its placement inside a {@link ReuseLibrary}.
	 *
	 * @param element        The target reuse element
	 * @param connectingSpec The reuse specification used as link/trace/reference
	 */
	public static void setupElementForLibrary(IModelElement element, ReuseElementSpec connectingSpec) {
		// currently, it is needed that only the original specification for this library
		// should exist (and also only the newest one) -> remove all others
		for (ReuseElementSpec spec : getAllReuseSpecs(element)) {
			if (!(connectingSpec.getElementUUID().equals(spec.getElementUUID())
					&& connectingSpec.getLastUpdate().equals(spec.getLastUpdate()))) {
				element.getSpecifications().remove(spec);
			}
		}
	}

	/**
	 * Creates a reuse element based on a given {@link EObject} for its placement
	 * inside a {@link ReuseLibrary}.
	 *
	 * @param originalElement The original element that should be used as base for
	 *                        the new reuse element
	 * @param library         The reuse library in which the new reuse element
	 *                        should be added (later)
	 * @param saveName        The name of the new reuse element with which it should
	 *                        be saved in the library
	 *
	 * @return The created reuse element for the library
	 */
	public static EObject createElementForLibrary(EObject originalElement, ReuseLibrary library, String saveName) {
		// add to the original element a reuse reference to the new library
		// (cast is safe: elements within the ReuseElementList are always IModelElement
		// by design)
		ReuseElementSpec spec = addReuseSpecToElement((IModelElement) originalElement, library);
		spec.setElementName(saveName);
		fixIds(originalElement);

		// copy, because origin should still exist further (but change name of copy)
		EObject newElementForLibrary = copy(originalElement);
		setReuseElementName(newElementForLibrary, saveName);
		setupElementForLibrary((IModelElement) newElementForLibrary, spec);

		return newElementForLibrary;
	}

	/**
	 * Adds a given element to the given {@link ReuseLibrary}.
	 *
	 * @param element The reuse element that should be added
	 * @param library The target reuse library
	 * @throws IOException
	 * @throws CoreException
	 */
	public static void addElementToLibrary(EObject element, ReuseLibrary library) throws IOException, CoreException {
		library.getReuseElementList().add(element);
		saveReuseLibrary(library);
	}

	/**
	 * Returns all {@link ReuseLibrary}s that are present in the local workspace.
	 * Will return empty list if no {@link ReuseLibrary} exists.
	 *
	 * @return A list of all local reuse libraries (or empty list)
	 */
	public static List<ReuseLibrary> getAllLocalReuseLibraries() {
		// currently, this is just all existing local libraries which are stored
		// separately in the library directory
		return getAllSeparatelyStoredReuseLibraries();
	}

	/**
	 * Returns the {@link ReuseLibrary} with the given UUID out of all local
	 * {@link ReuseLibrary}. Will return null if no {@link ReuseLibrary} exists with
	 * this UUID. Note: it will return the {@link ReuseLibrary} even if the folder
	 * or file name is not correct (as long as it is in the reuse directory and has
	 * the right UUID)!
	 *
	 * @param libraryUUID The UUID of the requested reuse library
	 * @return The requested reuse library (or null)
	 */
	public static ReuseLibrary getLocalReuseLibraryByID(String libraryUUID) {
		for (ReuseLibrary library : getAllLocalReuseLibraries()) {
			if (libraryUUID.equals(library.getLibraryUUID())) {
				return library;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link ReuseLibrary} that is mentioned as source library in the
	 * first {link ReuseElementSpec} of the given reuse element. If no {link
	 * ReuseElementSpec} exists, the return will be null. This method is faster than
	 * calling getAllSourceLibrariesOfElement() and returning the first library,
	 * because it does not fetch all libraries - only the one of the first {link
	 * ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 * @return The requested reuse library (or null)
	 */
	public static ReuseLibrary getFirstSourceLibraryOfElement(EObject element) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specList = getAllReuseSpecs((IModelElement) element);
			if (specList != null && !specList.isEmpty()) {
				// get(0) is safe due to check above
				String libraryUUID = specList.get(0).getSourceLibUUID();
				return getLocalReuseLibraryByID(libraryUUID);
			}
		}
		return null;
	}

	/**
	 * Returns a list of {@link ReuseLibrary}s that are each mentioned as source
	 * library in the {link ReuseElementSpec}s of the given reuse element. If no
	 * {link ReuseElementSpec} exists, the return will be an empty list.
	 *
	 * @param element The target reuse element
	 * @return All found reuse libraries for the given element (or empty list)
	 */
	public static List<ReuseLibrary> getAllSourceLibrariesOfElement(EObject element) {
		List<ReuseLibrary> foundLibraries = new ArrayList<>();
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> reuseSpecs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec reuseSpec : reuseSpecs) {
				String libraryUUID = reuseSpec.getSourceLibUUID();
				ReuseLibrary sourceLibrary = getLocalReuseLibraryByID(libraryUUID);
				if (sourceLibrary != null) {
					foundLibraries.add(sourceLibrary);
				}
			}
		}
		return foundLibraries;
	}

	/**
	 * Creates a new {@link ReuseLibrary} with the given name. If no name is
	 * received, the default library name will be used.
	 *
	 * @param libraryName The name of the new reuse library
	 * @return The requested reuse library
	 */
	public static ReuseLibrary createAndAddReuseLibrary(String libraryName) {
		if (libraryName == null || libraryName.equals("")) {
			return createReuseLibrary();
		} else {
			return createReuseLibrary(libraryName);
		}
	}

	/**
	 * Returns the element that exists in the given {@link ReuseLibrary} and is
	 * referenced by the given {@link ReuseElementSpec}. Returns null if it could
	 * not be found.
	 *
	 * @param library     The target reuse library
	 * @param elementUUID The UUID string of the requested element
	 * @return The requested reuse element (or null)
	 */
	public static EObject getElementInsideLibrary(ReuseLibrary library, String elementUUID) {
		// find original element inside the library by identical UUID
		for (EObject libraryElement : library.getReuseElementList()) {
			for (IModelElementSpecification spec : ((IModelElement) libraryElement).getSpecifications()) {
				if (spec instanceof ReuseElementSpec) {
					if (((ReuseElementSpec) spec).getElementUUID().equals(elementUUID)) {
						return libraryElement;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Returns a list of external elements that are used within the given element
	 * (e.g. an external stored function inside a code specification of a
	 * component). If nothing was found, an empty list is returned.
	 *
	 * @param element The target reuse element
	 * @return A list of external elements referencing the target reuse element (or
	 *         empty list)
	 */
	public static List<EObject> getExternalReferencesOfElement(EObject element) {
		List<EObject> references = new ArrayList<>();
		for (Class<? extends EObject> referenceClass : getPossibleExternalReferenceClasses()) {
			for (EObject reference : getChildrenWithType(element, referenceClass)) {
				references.add(reference);
			}
		}
		return references;
	}

	/**
	 * Replaces the given target/original reuse element in the given
	 * {@link ReuseLibrary} with the given new reuse element (which should have the
	 * given name in the end). The {@link ReuseElementSpec} references the original
	 * element (with its library).
	 *
	 * @param originalElementInLibrary The target reuse element inside a reuse
	 *                                 library that should be replace by the new
	 *                                 reuse element
	 * @param newElement               The new reuse element which is the
	 *                                 replacement of the current element in the
	 *                                 given library
	 * @param referencingSpec          The reuse specification that points to the
	 *                                 original reuse element in the library
	 * @param newName                  The new name of the element after replacement
	 */
	public static void replaceElementInLibrary(EObject originalElementInLibrary, EObject newElement,
			ReuseElementSpec referencingSpec, String newName) {
		// first, the information of the reference and the new element
		// needs to be updated and set up
		Date now = new Date();
		referencingSpec.setLastUpdate(now);
		referencingSpec.setElementName(newName);
		referencingSpec.setElementHash(createReuseHashAsString(newElement, now));

		EObject newLibraryElement = copy(newElement);
		setReuseElementName(newLibraryElement, newName);
		setupElementForLibrary((IModelElement) newLibraryElement, referencingSpec);

		replace(originalElementInLibrary, newLibraryElement);
	}

	/**
	 * Replaces the given (reuse) element inside its project with the given
	 * (original) reuse element from a {@link ReuseLibrary} (the element should have
	 * the given name in the end). Returns the new element in the project if the
	 * replacement/update was successful, otherwise null.
	 *
	 * @param elementInProject         The new reuse element which is the
	 *                                 replacement of the current element in the
	 *                                 given library
	 * @param originalElementInLibrary The target reuse element inside a reuse
	 *                                 library that should be replace by the new
	 *                                 reuse element
	 * @param newName                  The new name of the element after replacement
	 *
	 * @return The new element in the project if the replacement/update was
	 *         successful, otherwise null.
	 */
	public static EObject replaceElementInProject(EObject elementInProject, EObject originalElementInLibrary,
			String newName) {
		// first, the information of the reference and the new element
		// needs to be updated and set up
		List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) originalElementInLibrary);
		if (specs.size() == 1) {
			// get(0) is safe due to size check
			specs.get(0).setLastUpdate(new Date());
		} else {
			return null;
		}

		EObject sourceElement = copy(originalElementInLibrary);
		setReuseElementName(sourceElement, newName);

		boolean success = IReuseProviderService.getInstance().updateReuseElement(elementInProject, sourceElement);
		return success ? sourceElement : null;
	}

	/**
	 * Deletes the given {@link ReuseLibrary}.
	 *
	 * @param library The target reuse library that should be deleted
	 */
	public static void deleteReuseLibrary(ReuseLibrary library) {
		// currently, everything is based only on the reuse folder/file structure, which
		// is why it is enough to just delete the local folder
		try {
			deleteReuseLibraryFolder(library);
		} catch (CoreException e) {
			error(getDefault(), "Error during deleting reuse library folder", e);
			showError("Cannot delete reuse library: " + getReuseElementName(library));
		}
	}

	/**
	 * Deletes the given reuse element by deleting its entry in the local file of
	 * the given {@link ReuseLibrary} (if it exists there).
	 *
	 * @param element       The target reuse element that should be deleted
	 * @param sourceLibrary The reuse library in which the deletion should happen
	 */
	public static void deleteReuseElementInLibraryFile(EObject element, ReuseLibrary sourceLibrary) {
		// the objects might not be the same anymore, which is why we need to identify
		// the correct object via the reuse UUID
		EObject correctLibraryElement = null;
		EList<EObject> libraryElementList = sourceLibrary.getReuseElementList();
		for (EObject libraryElement : libraryElementList) {
			if (getFirstReuseElementUUID(libraryElement)
					.equals(getReuseElementUUIDForLibrary(element, sourceLibrary))) {
				correctLibraryElement = libraryElement;
				break;
			}
		}
		if (correctLibraryElement == null) {
			return;
		}
		libraryElementList.remove(correctLibraryElement);

		try {
			saveReuseLibrary(sourceLibrary);
		} catch (IOException | CoreException e) {
			error(getDefault(), "Error during removing reuse element from reuse library", e);
			showError("Cannot remove reuse element: " + getReuseElementName(element) + " from this reuse library: "
					+ getReuseElementName(sourceLibrary));
		}
	}

	/**
	 * Deletes the given reuse element by deleting its entry in the local file of
	 * the {@link ReuseLibrary} that is mentioned as first (or only) source library
	 * for this reuse element (in its {link ReuseElementSpec}s).
	 *
	 * @param element The target reuse element that should be deleted
	 */
	public static void deleteReuseElementInLibraryFile(EObject element) {
		ReuseLibrary sourceLibrary = getFirstSourceLibraryOfElement(element);
		if (sourceLibrary != null) {
			deleteReuseElementInLibraryFile(element, sourceLibrary);
		}
	}

	/**
	 * Updates the name directly in the file of the given reuse element, which can
	 * also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), the given source
	 * {@link ReuseLibrary} is where the reuse element will be updated with the new
	 * name.
	 *
	 * @param element       The target reuse element that should be updated
	 * @param newName       The new name as string
	 * @param sourceLibrary The reuse library in which the update should happen
	 */
	public static void updateReuseElementFileWithName(EObject element, String newName, ReuseLibrary sourceLibrary) {
		if (element == null || newName == null || newName.isBlank()) {
			return;
		}

		boolean modifiedLibrary = false;
		ReuseLibrary libraryForDeletion = null;
		String errorObject = "";
		if (element instanceof ReuseLibrary) {
			// since the folder and file names of a reuse library contain the library name
			// (beside its UUID), the easiest way to rename it is to save it as new library
			// (where only the name is changed, not UUID, etc.) and delete its previous
			// library folder (save and deletion are done at the end of this method)
			sourceLibrary = (ReuseLibrary) element;
			// copy, because origin should still exist further for deletion
			libraryForDeletion = copy(sourceLibrary);

			setReuseElementName(sourceLibrary, newName);
			for (EObject libraryElement : sourceLibrary.getReuseElementList()) {
				// cast is safe: elements within the ReuseElementList are always IModelElement
				// by design
				List<ReuseElementSpec> specList = getAllReuseSpecs((IModelElement) libraryElement);
				for (ReuseElementSpec reuseSpec : specList) {
					reuseSpec.setSourceLibName(newName);
				}
			}

			errorObject = "reuse library '" + getReuseElementName(sourceLibrary) + "'";
			modifiedLibrary = true;
		} else {
			// the objects might not be the same anymore, which is why we need to identify
			// the correct object via the reuse UUID
			for (EObject libraryElement : sourceLibrary.getReuseElementList()) {
				if (getFirstReuseElementUUID(libraryElement)
						.equals(getReuseElementUUIDForLibrary(element, sourceLibrary))) {
					setReuseElementName(libraryElement, newName);
					// cast is safe: elements within the ReuseElementList are always IModelElement
					// by design
					List<ReuseElementSpec> specList = getAllReuseSpecs((IModelElement) libraryElement);
					for (ReuseElementSpec reuseSpec : specList) {
						reuseSpec.setElementName(newName);
					}
					setReuseElementLastUpdateNow(libraryElement);

					errorObject = "reuse element '" + getReuseElementName(libraryElement) + "' in reuse library '"
							+ getReuseElementName(sourceLibrary) + "'";
					modifiedLibrary = true;
					break;
				}
			}
		}

		if (modifiedLibrary) {
			try {
				saveReuseLibrary(sourceLibrary);
				if (libraryForDeletion != null) {
					deleteReuseLibraryFolder(libraryForDeletion);
				}
			} catch (IOException | CoreException e) {
				error(getDefault(), "Error during renaming " + errorObject, e);
				showError("Cannot rename " + errorObject);
			}
		}
	}

	/**
	 * Updates the name directly in the file of the given reuse element, which can
	 * also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), its first (or only) source
	 * {@link ReuseLibrary}, which is mentioned in its {link ReuseElementSpec}s, is
	 * where the reuse element will be updated with the new name.
	 *
	 * @param element The target reuse element that should be updated
	 * @param newName The new name as string
	 */
	public static void updateReuseElementFileWithName(EObject element, String newName) {
		if (element instanceof ReuseLibrary) {
			updateReuseElementFileWithName(element, newName, null);
		} else {
			ReuseLibrary sourceLibrary = getFirstSourceLibraryOfElement(element);
			if (sourceLibrary != null) {
				updateReuseElementFileWithName(element, newName, sourceLibrary);
			}
		}
	}

	/**
	 * Updates the comment directly in the file of the given reuse element, which
	 * can also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), the given source
	 * {@link ReuseLibrary} is where the reuse element will be updated with the new
	 * comment.
	 *
	 * @param element       The target reuse element that should be updated
	 * @param newComment    The new comment as string
	 * @param sourceLibrary The reuse library in which the update should happen
	 */
	public static void updateReuseElementFileWithComment(EObject element, String newComment,
			ReuseLibrary sourceLibrary) {
		boolean modifiedLibrary = false;
		String errorObject = "";
		if (element instanceof ReuseLibrary) {
			sourceLibrary = (ReuseLibrary) element;
			setReuseElementComment(sourceLibrary, newComment);

			errorObject = "reuse library '" + getReuseElementName(sourceLibrary) + "'";
			modifiedLibrary = true;
		} else {
			// the objects might not be the same anymore, which is why we need to identify
			// the correct object via the reuse UUID
			EList<EObject> libraryElementList = sourceLibrary.getReuseElementList();
			for (EObject libraryElement : libraryElementList) {
				if (getFirstReuseElementUUID(libraryElement)
						.equals(getReuseElementUUIDForLibrary(element, sourceLibrary))) {
					setReuseElementComment(libraryElement, newComment);
					setReuseElementLastUpdateNow(libraryElement);

					errorObject = "reuse element '" + getReuseElementName(libraryElement) + "' in reuse library '"
							+ getReuseElementName(sourceLibrary) + "'";
					modifiedLibrary = true;
					break;
				}
			}
		}

		if (modifiedLibrary) {
			try {
				saveReuseLibrary(sourceLibrary);
			} catch (IOException | CoreException e) {
				error(getDefault(), "Error during modifying " + errorObject, e);
				showError("Cannot modify " + errorObject);
			}
		}
	}

	/**
	 * Updates the comment directly in the file of the given reuse element, which
	 * can also be a {@link ReuseLibrary}, with the given {link String}. If it is an
	 * actual reuse element (not a {@link ReuseLibrary}), its first (or only) source
	 * {@link ReuseLibrary}, which is mentioned in its {link ReuseElementSpec}s, is
	 * where the reuse element will be updated with the new comment.
	 *
	 * @param element    The target reuse element that should be updated
	 * @param newComment The new comment as string
	 */
	public static void updateReuseElementFileWithComment(EObject element, String newComment) {
		if (element instanceof ReuseLibrary) {
			updateReuseElementFileWithComment(element, newComment, null);
		} else {
			ReuseLibrary sourceLibrary = getFirstSourceLibraryOfElement(element);
			if (sourceLibrary != null) {
				updateReuseElementFileWithComment(element, newComment, sourceLibrary);
			}
		}
	}

	/**
	 * Creates a hash code for the given {@link EObject} as reuse element and
	 * returns it as {@link String}. This has code format is used for detecting
	 * changes between reuse elements with the same origin. Returns null if given
	 * element is invalid (e.g. not a reuse element).
	 *
	 * @param element    The element of which a hash code should be created
	 * @param modifyDate The date of the last update/modification of the given
	 *                   element
	 * @return String of the created hash code integer (or null for an invalid
	 *         input)
	 */
	public static String createReuseHashAsString(EObject element, Date modifyDate) {
		if (element != null && modifyDate != null) {
			// hashCode() does not change through element differences caused by simple
			// changes like name or new ports, etc.
			// Therefore, we are currently modifying the hash with the date of the last
			// update to be sure that the hash is different after a reuse update (to see
			// e.g. if the original reuse element in the library was updated by someone else
			// in the meantime).
			int hash = element.hashCode();
			String datePattern = "yyyyMMddHHmmss";
			DateFormat dateFormat = new SimpleDateFormat(datePattern);
			String dateAsString = dateFormat.format(modifyDate);
			String hashString = String.valueOf(hash) + dateAsString;
			return hashString;
		}
		return null;
	}

	/**
	 * Checks whether the given {@link EObject} has an origin in a
	 * {@link ReuseLibrary} and then returns true if the origin has a different hash
	 * code than the (given) reuse element (meaning that its origin in the library
	 * was updated since its reuse). If the given {@link EObject} has no connection
	 * to a {@link ReuseLibrary} or the original element in the library is still
	 * identical, the function will return false.
	 *
	 * @param element The element whose reuse origin should be checked for updates
	 * @return True if the given element has a reuse origin and this origin was
	 *         updated in the meantime, otherwise false
	 */
	public static List<EObject> getAllUpdatedOrigins(EObject element) {
		// TODO: use this method for detecting updates and informing user (separate
		// context menu entry?)
		List<EObject> updatedOrigins = new ArrayList<>();

		List<ReuseLibrary> foundLibraries = getAllSourceLibrariesOfElement(element);

		for (ReuseLibrary sourceLibrary : foundLibraries) {
			System.out.println("Lib: " + sourceLibrary);
			String uuid = getReuseElementUUIDForLibrary(element, sourceLibrary);
			EObject originElementInLib = getElementInsideLibrary(sourceLibrary, uuid);
			String storedHashForThisLibrary = getReuseElementHashForLibrary(element, sourceLibrary);
			String currentHashForThisLibrary = getReuseElementHashForLibrary(originElementInLib, sourceLibrary);
			// if hashes are valid but different, the origin was updated
			System.out.println("Hashes: " + storedHashForThisLibrary + " vs " + currentHashForThisLibrary + "\n");
			if (!storedHashForThisLibrary.isEmpty() && !currentHashForThisLibrary.isEmpty()) {
				if (!storedHashForThisLibrary.equals(currentHashForThisLibrary)) {
					updatedOrigins.add(originElementInLib);
				}
			}
		}
		return updatedOrigins;
	}

}
